import logging
import os
import platform
import signal
from datetime import datetime
from logging import Logger
from typing import Optional
from uuid import uuid4


def force_kill_process(pid: int):
    system = platform.system()
    print(f"Force killing {pid=}")
    if system == "Windows":
        os.system(f"taskkill /f /pid {pid}")
    else:
        os.kill(pid, signal.SIGKILL)


def generate_id(time: Optional[datetime] = None, uuid_length: int = 6):
    time = time or datetime.now()
    time_part = time.strftime("%H%M_%d%m_")
    uuid_part = str(uuid4()).replace("-", "")[:uuid_length]
    return time_part + uuid_part


def setup_logger():
    logger = Logger(name=__name__, level=logging.INFO)
    handler = logging.StreamHandler()
    formatter = logging.Formatter('%(levelname)s:%(process)d:%(message)s')
    handler.setLevel(level=logging.INFO)
    handler.setFormatter(fmt=formatter)
    logger.addHandler(hdlr=handler)
    return logger



from copy import copy
from multiprocessing import Lock, Value, Queue
from ctypes import c_int, c_bool


class Controller:
    default_timeout_sec = 10

    def __init__(self):
        self._lock: Lock = Lock()
        self._stop_flag = Value(c_bool, False)
        self._status = Value(c_int, 0)
        self._input = Queue()
        self._output = Queue()

    def set_stop(self, value: bool = True):
        with self._lock:
            self._stop_flag.value = value

    def get_stop(self):
        with self._lock:
            return copy(self._stop_flag.value)

    def set_status(self, value: int):
        with self._lock:
            self._status.value = value

    def get_status(self):
        with self._lock:
            return copy(self._status.value)

    def insert_input(self, obj):
        self._insert_queue(queue=self._input, obj=obj)

    def pop_input(self):
        return self._pop_queue(queue=self._input)

    def insert_output(self, obj):
        self._insert_queue(queue=self._output, obj=obj)

    def pop_output(self):
        return self._pop_queue(queue=self._output)

    def _insert_queue(self, queue: Queue, obj):
        with self._lock:
            queue.put(obj=obj, block=True, timeout=self.default_timeout_sec)

    def _pop_queue(self, queue: Queue):
        with self._lock:
            return queue.get(block=True, timeout=self.default_timeout_sec)



from logging import getLogger, Logger
from time import sleep, time
from typing import Optional, Callable, Any

from service_sdk.exceptions import TimeoutWaitingForWorkerToStop
from service_sdk.manager.controller import Controller
from service_sdk.manager.registry import Registry
from service_sdk.worker.worker import Worker


class WorkManager:
    def __init__(
            self,
            name=None,
            logger: Optional[Logger] = None,
            work_callback: Optional[Callable[[Any], Any]] = None,
            max_workers: int = 2,
    ):
        self.name = name
        self.logger = logger or getLogger(__name__)
        self.work_callback = work_callback or (lambda **_: None)
        self.max_workers = max_workers
        self._registry = Registry(max_workers=max_workers)

    def start_worker(self, **kwargs):
        self._update_workers_status()
        self.logger.info("Starting worker")

        if self._registry.counter >= self.max_workers:
            self.logger.info("Max worker capacity reached.")
            return

        worker = Worker(controller=Controller(), work_callback=self.work_callback)
        self._registry.add(worker=worker)
        worker.start()
        self.logger.info(f"Worker started. uid={worker.uid}, pid={worker.get_pid()}.")

    def stop_all(self, wait: bool = True, timeout: int = 10):
        for worker in self._registry.all():
            worker.controller.set_stop()

        if wait:
            self._wait_for_workers_to_stop(timeout=timeout)

    def stop_by_uid(self, uid: str) -> Optional[str]:
        worker = self._registry.get(id_=uid)
        if not worker:
            return
        worker.controller.set_stop()
        self._update_workers_status()
        return worker.uid

    def kill_all(self, wait: bool = True, timeout: int = 10):
        for worker in self._registry.all():
            worker.process.kill()

        if wait:
            self._wait_for_workers_to_stop(timeout=timeout)

    def _wait_for_workers_to_stop(self, timeout: int = 10):
        start_time = time()
        while self._registry.counter > 0:
            if start_time - time() > timeout:
                self.logger.error(
                    f"Reached {timeout=} while waiting for workers to stop. {self._registry.counter} left."
                )
                raise TimeoutWaitingForWorkerToStop()

            self._update_workers_status()
            self.logger.info(f"Waiting for workers to stop. {self._registry.counter} left.")
            sleep(0.5)

    def get_status(self):
        self._update_workers_status()
        status = {}
        for worker in self._registry.all():
            status[worker.uid] = worker.controller.get_status()
        return status

    def restart(self):
        self.stop_all()
        self.start_worker()

    def _update_workers_status(self):
        for worker in self._registry.all():
            if not worker.process.is_alive():
                self._registry.remove(id_=worker.uid)



from multiprocessing import Process
from time import sleep
from typing import Callable, Any, Optional

from service_sdk.exceptions import StopGracefullyException
from service_sdk.manager.controller import Controller
from service_sdk.utils import generate_id, setup_logger
from service_sdk.worker.ex_thread import ExThread


class Worker:
    def __init__(
            self,
            controller: Controller,
            work_callback: Callable[[Any], Any],
            uid: Optional[str] = None,
            **kwargs
    ):
        self.work_callback = work_callback
        self.controller = controller
        self.uid = uid or generate_id()
        self.process = Process(target=self.handle_work, kwargs=kwargs)
        self.context = {"id": self.uid}

    def get_pid(self) -> Optional[int]:
        return self.process.pid

    def start(self):
        self.process.start()

    def handle_work(self, **kwargs):
        logger = setup_logger()
        counter = 0
        logger.info("Starting work")

        working_thread = ExThread(target=self.work_callback, daemon=True, kwargs=dict(context=self.context))
        working_thread.start()
        while not self.controller.get_stop():
            counter += 1
            self.controller.set_status(counter)
            sleep(1)

        logger.info("Stop command received. Shutting worker down.")
        working_thread.raise_exception(exception=StopGracefullyException)
